/*
 * deviced
 *
 * Copyright (c) 2019 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "core/devices.h"
#include "core/common.h"
#include "core/device-notifier.h"
#include "display/util.h"
#include "display/poll.h"
#include "display-ops.h"
#include <hw/bezel.h>
#include <assert.h>
#include <vconf.h>

#ifndef VCONFKEY_SETAPPL_WAKEUP_BY_BEZEL_ENABLE
#define VCONFKEY_SETAPPL_WAKEUP_BY_BEZEL_ENABLE "db/setting/wakeup_by_bezel_enable"
#endif

static struct bezel_device *bezel_dev;
static int bezel_wakeup = -1;

static void bezel_changed_cb(keynode_t *key_nodes, void *data)
{
	assert(key_nodes);

	bezel_wakeup = vconf_keynode_get_bool(key_nodes);

	_I("bezel wakeup condition is %d", bezel_wakeup);
	if (disp_plgn.pm_change_internal)
		disp_plgn.pm_change_internal(INTERNAL_LOCK_PM, LCD_NORMAL);

	device_notify(DEVICE_NOTIFIER_BEZEL_WAKEUP, (void *)bezel_wakeup);
}

static void bezel_init(void *data)
{
	struct hw_info *info;
	int r;

	if (bezel_dev)
		return;

	r = hw_get_info(BEZEL_HARDWARE_DEVICE_ID, (const struct hw_info **)&info);
	if (r < 0) {
		_I("bezel shared library is not supported: %d", r);
		return;
	}

	if (!info->open) {
		_E("Fail to open bezel device : open(NULL)");
		return;
	}

	r = info->open(info, NULL, (struct hw_common **)&bezel_dev);
	if (r < 0) {
		_E("Fail to get bezel device structure: %d", r);
		return;
	}

	vconf_notify_key_changed(VCONFKEY_SETAPPL_WAKEUP_BY_BEZEL_ENABLE,
			bezel_changed_cb, NULL);

	r = vconf_get_bool(VCONFKEY_SETAPPL_WAKEUP_BY_BEZEL_ENABLE, &bezel_wakeup);
	if (r < 0) {
		_E("Failed to get vconf value for wakeup by bezel enable: %d", vconf_get_ext_errno());
		bezel_wakeup = 1;
	}

	_I("bezel wakeup condition is %d", bezel_wakeup);
}

static void bezel_exit(void *data)
{
	struct hw_info *info;

	if (!bezel_dev)
		return;

	info = bezel_dev->common.info;

	assert(info);

	info->close((struct hw_common *)bezel_dev);
}

static int bezel_start(enum device_flags flags)
{
	if (!bezel_dev)
		return 0;

	return bezel_dev->set_state(BEZEL_TURNON);
}

static int bezel_stop(enum device_flags flags)
{
	enum bezel_state state;

	if (!bezel_dev)
		return 0;

	state = (bezel_wakeup ? BEZEL_TURNON : BEZEL_TURNOFF);
	_I("set bezel state: %d", state);

	return bezel_dev->set_state(state);
}

/*
 * 'bezel' device_ops should be initialize before init key-filter.
 */
static const struct device_ops bezel_device_ops = {
	.priority = DEVICE_PRIORITY_HIGH,
	.name     = "bezel",
	.init     = bezel_init,
	.exit     = bezel_exit,
	.start    = bezel_start,
	.stop     = bezel_stop,
};

DEVICE_OPS_REGISTER(&bezel_device_ops)
