/*
 * deviced
 *
 * Copyright (c) 2012 - 2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <stdbool.h>
#include <vconf.h>
#include <device-node.h>
#include <bundle.h>
#include <eventsystem.h>
#include <hw/battery.h>

#include "core/devices.h"
#include "core/device-notifier.h"
#include "core/udev.h"
#include "core/log.h"
#include "core/config-parser.h"
#include "display/poll.h"
#include "display/setting.h"
#include "display/display-ops.h"
#include "apps/apps.h"
#include "power-supply.h"
#include "battery.h"

#define BATTERY_NAME        "battery"
#define CHARGEFULL_NAME     "Full"
#define CHARGENOW_NAME      "Charging"
#define DISCHARGE_NAME      "Discharging"
#define NOTCHARGE_NAME      "Not charging"
#define OVERHEAT_NAME       "Overheat"
#define TEMPCOLD_NAME       "Cold"
#define OVERVOLT_NAME       "Over voltage"
#define GOOD_NAME           "Good"

#define REMOVE_POPUP "remove_battery_popups"

#define BUFF_MAX            255

#define SIGNAL_CHARGEERR_RESPONSE "ChargeErrResponse"
#define SIGNAL_TEMP_GOOD          "TempGood"

#define ABNORMAL_CHECK_TIMER_INTERVAL 60

#define METHOD_FULL_NOTI_ON   "BatteryFullNotiOn"
#define METHOD_FULL_NOTI_OFF  "BatteryFullNotiOff"
#define METHOD_CHARGE_NOTI_ON "BatteryChargeNotiOn"

#define RETRY_MAX 5
#define BATTERY_CHECK_TIMER_INTERVAL 500  /* 0.5 second */

enum power_supply_init_type {
	POWER_SUPPLY_NOT_READY   = 0,
	POWER_SUPPLY_INITIALIZED = 1,
};

struct battery_status battery;
struct battery_status old_battery;
static int noti_id;
static guint power_timer;

static int booting_done(void *data);

static struct battery_device *battery_dev;

static void lowbat_execute(void *data)
{
	static const struct device_ops *lowbat_ops;

	FIND_DEVICE_VOID(lowbat_ops, "lowbat");
	device_execute(lowbat_ops, data);
}

static void pm_check_and_change(int bInserted)
{
	static int old = -1;

	if (old == bInserted)
		return;
	old = bInserted;
	if (disp_plgn.pm_change_internal)
		disp_plgn.pm_change_internal(INTERNAL_LOCK_POPUP, LCD_NORMAL);
}

static int changed_battery_cf(enum present_type status)
{
	char *value;

	if (status == PRESENT_ABNORMAL)
		value = "battdisconnect";
	else
		value = "remove_battery_popups";

	return launch_system_app(APP_DEFAULT,
			2, APP_KEY_TYPE, value);
}

static void health_status_broadcast(void)
{
	dbus_handle_broadcast_dbus_signal(DEVICED_PATH_BATTERY, DEVICED_INTERFACE_BATTERY,
	    SIGNAL_TEMP_GOOD, NULL, NULL);
}

#ifdef TIZEN_FEATURE_BATTERY_OVER_TEMPERATURE
static guint abnormal_timer;

static void abnormal_popup_timer_init(void)
{
	if (abnormal_timer == 0)
		return;

	g_source_remove(abnormal_timer);
	abnormal_timer = 0;
	_I("Delete health timer.");
}

static void health_timer_reset(void)
{
	abnormal_timer = 0;
}

static gboolean health_timer_cb(void *data)
{
	int value = HEALTH_DEAD;

	health_timer_reset();

	if (battery.health == HEALTH_GOOD)
		return G_SOURCE_REMOVE;

	_I("Popup: Battery health status is not good.");
	device_notify(DEVICE_NOTIFIER_BATTERY_HEALTH, (void *)&value);
	if (disp_plgn.pm_change_internal)
		disp_plgn.pm_change_internal(INTERNAL_LOCK_POPUP, LCD_NORMAL);
	if (disp_plgn.pm_lock_internal)
		disp_plgn.pm_lock_internal(INTERNAL_LOCK_POPUP, LCD_DIM, STAY_CUR_STATE, 0);
	if (battery.health == HEALTH_LOW)
		battery_charge_err_low_act(NULL);
	else if (battery.health == HEALTH_HIGH)
		battery_charge_err_high_act(NULL);
	return G_SOURCE_REMOVE;
}

static void abnormal_popup_dbus_signal_handler(GDBusConnection  *conn,
		const gchar      *sender,
		const gchar      *path,
		const gchar      *iface,
		const gchar      *name,
		GVariant         *param,
		gpointer          user_data)

{
	if (battery.health == HEALTH_GOOD)
		return;
	_I("Restart health timer.");
	abnormal_timer = g_timeout_add_seconds(ABNORMAL_CHECK_TIMER_INTERVAL,
						health_timer_cb, NULL);
	if (abnormal_timer == 0)
		_E("Failed to add abnormal check timer.");
}
#endif /* TIZEN_FEATURE_BATTERY_OVER_TEMPERATURE */

static void full_noti_cb(GVariant *var, void *user_data, GError *err)
{
	int id = 0;

	if (!var)
		return;

	if (!dh_get_param_from_var(var, "(i)", &id)) {
		_E("Failed to notify full: no message(%s)", g_variant_get_type_string(var));
		goto out;
	}

	noti_id = id;
	_D("Inserted battery full noti(%d).", noti_id);

out:
	g_variant_unref(var);
}

static void noti_off_cb(GVariant *var, void *user_data, GError *err)
{
	int ret = 0;

	if (!dh_get_param_from_var(var, "(i)", &ret)) {
		_E("Failed to off notification: no message(%s)", g_variant_get_type_string(var));
		goto out;
	}

	_D("Noti off: %d", ret);

out:
	g_variant_unref(var);
}

static int check_power_supply_noti(void)
{
	int block, ret;

	ret = vconf_get_bool("db/setting/blockmode_wearable", &block);
	if (ret < 0) {
		_E("Failed to get vconf value for blockmode wearable: %d", vconf_get_ext_errno());
		return 1;
	}
	if (block == 0)
		return 1;
	return 0;
}

static int send_full_noti(enum charge_full_type state)
{
	int ret = 0;
	int noti;
	int retry;

	noti = check_power_supply_noti();

	if (!noti)
		return noti;

	switch (state) {
	case CHARGING_FULL:
		for (retry = RETRY_MAX; retry > 0; retry--) {
			ret = dbus_handle_method_async_with_reply(POPUP_BUS_NAME,
					POPUP_PATH_NOTI,
					POPUP_INTERFACE_NOTI,
					METHOD_FULL_NOTI_ON,
					NULL, NULL, full_noti_cb, -1, NULL);
			if (ret == 0) {
				_D("Created battery full noti.");
				return ret;
			}
		}
		_E("Failed to call dbus method: %d", ret);
	break;
	case CHARGING_NOT_FULL:
		if (noti_id <= 0)
			return -EPERM;
		for (retry = RETRY_MAX; retry > 0; retry--) {
			ret = dbus_handle_method_async_with_reply_var(POPUP_BUS_NAME,
					POPUP_PATH_NOTI,
					POPUP_INTERFACE_NOTI,
					METHOD_FULL_NOTI_OFF,
					g_variant_new("(i)", noti_id),
					noti_off_cb, -1, NULL);
			if (ret == 0) {
				_D("Deleted battery full noti.");
				noti_id = 0;
				return ret;
			}
		}
		_E("Failed to call dbus method: %d", ret);
	break;
	}
	return ret;
}

static void charge_noti_on(GVariant *var, void *user_data, GError *err)
{
	int id = 0;

	if (!var)
		return;

	if (!dh_get_param_from_var(var, "(i)", &id)) {
		_E("Failed to notify charge: no message(%s)", g_variant_get_type_string(var));
		goto out;
	}

	_D("Inserted battery charge noti(%d)", id);

out:
	g_variant_unref(var);
}

static int send_charge_noti(void)
{
	int ret = 0;
	int retry;

	for (retry = RETRY_MAX; retry > 0; retry--) {
		ret = dbus_handle_method_async_with_reply(POPUP_BUS_NAME,
				POPUP_PATH_NOTI,
				POPUP_INTERFACE_NOTI,
				METHOD_CHARGE_NOTI_ON,
				NULL, NULL, charge_noti_on, -1, NULL);
		if (ret == 0) {
			_D("Created battery charge noti.");
			return ret;
		}
	}
	_E("Failed to call dbus method: %d", ret);
	return ret;
}

static void power_supply_noti(enum battery_noti_type type, enum battery_noti_status status)
{
	static int charger = CHARGER_DISCHARGING;
	static int full = CHARGING_NOT_FULL;
	int ret;

	if (type == DEVICE_NOTI_BATT_CHARGE) {
		if (status == DEVICE_NOTI_ON && charger == CHARGER_DISCHARGING) {
			send_charge_noti();
			charger = CHARGER_CHARGING;
		} else if (status == DEVICE_NOTI_OFF && charger == CHARGER_CHARGING) {
			charger = CHARGER_DISCHARGING;
		}
	} else if (type == DEVICE_NOTI_BATT_FULL) {
		if (status == DEVICE_NOTI_ON && full == CHARGING_NOT_FULL) {
			ret = send_full_noti(CHARGING_FULL);
			if (ret == 0)
				full = CHARGING_FULL;
		} else if (status == DEVICE_NOTI_OFF && full == CHARGING_FULL) {
			ret = send_full_noti(CHARGING_NOT_FULL);
			if (ret == 0)
				full = CHARGING_NOT_FULL;
		}
	}
}

void power_supply_broadcast(char *sig, int status)
{
	static int old;
	static char sig_old[32];

	if (strcmp(sig_old, sig) == 0 && old == status)
		return;

	_D("signal=%s status=%d", sig, status);

	old = status;
	snprintf(sig_old, sizeof(sig_old), "%s", sig);
	/*snprintf(str_status, sizeof(str_status), "%d", status);
	arr[0] = str_status;*/

	dbus_handle_broadcast_dbus_signal_var(DEVICED_PATH_BATTERY,
						DEVICED_INTERFACE_BATTERY,
						sig,
						g_variant_new("(i)", status));

}

static void noti_batt_full(void)
{
	static int bat_full_noti;
	int noti;

	if (!battery.charge_full && bat_full_noti == 1) {
		power_supply_noti(DEVICE_NOTI_BATT_FULL, DEVICE_NOTI_OFF);
		bat_full_noti = 0;
		/* off the full charge state */
		device_notify(DEVICE_NOTIFIER_FULLBAT, (void *)&bat_full_noti);
	}
	if (battery.charge_full && bat_full_noti == 0) {
		power_supply_noti(DEVICE_NOTI_BATT_FULL, DEVICE_NOTI_ON);
		bat_full_noti = 1;
		/* turn on LCD, if battery is full charged */
		noti = check_power_supply_noti();
		if (noti) {
			if (disp_plgn.pm_change_internal)
				disp_plgn.pm_change_internal(INTERNAL_LOCK_BATTERY_FULL, LCD_NORMAL);
		} else {
			_I("Block LCD.");
		}
		/* on the full charge state */
		device_notify(DEVICE_NOTIFIER_FULLBAT, (void *)&bat_full_noti);
	}
}

static void check_power_supply(int state)
{
	pm_check_and_change(state);
	if (disp_plgn.update_pm_setting)
		disp_plgn.update_pm_setting(SETTING_CHARGING, state);
}

static void charger_state_send_system_event(int state)
{
	bundle *b;
	const char *str;

	switch (state) {
	case CHARGE_STATUS_CHARGING:
		str = EVT_VAL_BATTERY_CHARGER_CHARGING;
		break;
	case CHARGE_STATUS_FULL:
	case CHARGE_STATUS_DISCHARGING:
		str = EVT_VAL_BATTERY_CHARGER_DISCHARGING;
		break;
	case CHARGE_STATUS_CONNECTED:
		str = EVT_VAL_BATTERY_CHARGER_CONNECTED;
		break;
	case CHARGE_STATUS_DISCONNECTED:
		str = EVT_VAL_BATTERY_CHARGER_DISCONNECTED;
		break;
	default:
		_E("Invalid parameter(%d).", state);
		return;
	}

	_D("system_event(%s)", str);

	b = bundle_create();
	bundle_add_str(b, EVT_KEY_BATTERY_CHARGER_STATUS, str);
	eventsystem_send_system_event(SYS_EVENT_BATTERY_CHARGER_STATUS, b);
	bundle_free(b);
}

static void update_present(enum battery_noti_status status)
{
	static int old = DEVICE_NOTI_OFF;
	enum present_type present;

	if (old == status)
		return;
	_I("charge=%d present=%d", battery.charge_now, battery.present);
	old = status;
	if (disp_plgn.pm_change_internal)
		disp_plgn.pm_change_internal(INTERNAL_LOCK_POPUP, LCD_NORMAL);
	if (status == DEVICE_NOTI_ON) {
		present = PRESENT_ABNORMAL;
		device_notify(DEVICE_NOTIFIER_BATTERY_PRESENT, (void *)&present);
		if (disp_plgn.pm_lock_internal)
			disp_plgn.pm_lock_internal(INTERNAL_LOCK_POPUP, LCD_DIM, STAY_CUR_STATE, 0);
	} else {
		present = PRESENT_NORMAL;
		device_notify(DEVICE_NOTIFIER_BATTERY_PRESENT, (void *)&present);
		if (disp_plgn.pm_unlock_internal)
			disp_plgn.pm_unlock_internal(INTERNAL_LOCK_POPUP, LCD_DIM, PM_SLEEP_MARGIN);
	}
	changed_battery_cf(present);
}

#ifdef TIZEN_FEATURE_BATTERY_OVER_TEMPERATURE
static void update_health(enum battery_noti_status status)
{
	static int old = DEVICE_NOTI_OFF;

	if (old == status)
		return;
	_I("charge=%d health=%d", battery.charge_now, battery.health);
	old = status;

	if (disp_plgn.pm_change_internal)
		disp_plgn.pm_change_internal(INTERNAL_LOCK_POPUP, LCD_NORMAL);

	if (status == DEVICE_NOTI_ON) {
		_I("Popup: Battery health status is not good.");

		if (disp_plgn.pm_lock_internal)
			disp_plgn.pm_lock_internal(INTERNAL_LOCK_POPUP, LCD_DIM, STAY_CUR_STATE, 0);
		if (battery.health == HEALTH_LOW)
			battery_charge_err_low_act(NULL);
		else if (battery.health == HEALTH_HIGH)
			battery_charge_err_high_act(NULL);
	} else {
		if (disp_plgn.pm_unlock_internal)
			disp_plgn.pm_unlock_internal(INTERNAL_LOCK_POPUP, LCD_DIM, PM_SLEEP_MARGIN);
		abnormal_popup_timer_init();
		launch_system_app(APP_DEFAULT, 2, APP_KEY_TYPE, REMOVE_POPUP);
	}
}
#endif /* TIZEN_FEATURE_BATTERY_OVER_TEMPERATURE */

static void update_ovp(enum battery_noti_status status)
{
	static int old = DEVICE_NOTI_OFF;
	int value;

	if (old == status)
		return;
	_I("charge=%d ovp=%d", battery.charge_now, battery.health);
	old = status;
	if (disp_plgn.pm_change_internal)
		disp_plgn.pm_change_internal(INTERNAL_LOCK_POPUP, LCD_NORMAL);
	if (status == DEVICE_NOTI_ON)
		value = HEALTH_OVP;
	else
		value = HEALTH_GOOD;
	device_notify(DEVICE_NOTIFIER_BATTERY_OVP, (void *)&value);
}

static void check_battery_status(void)
{
	static int old = DEVICE_CHANGE_NORMAL;
	int status;

	if (battery.charge_now == CHARGER_ABNORMAL &&
	    (battery.health == HEALTH_DEAD || battery.present == PRESENT_ABNORMAL))
		status = DEVICE_CHANGE_ABNORMAL;
	else if (battery.health == HEALTH_OVP)
		status = DEVICE_CHANGE_ABNORMAL;
	else
		status = DEVICE_CHANGE_NORMAL;
	if (old == status)
		return;
	old = status;

	if (battery.charge_now == CHARGER_ABNORMAL) {
		if (battery.health == HEALTH_DEAD) {
			device_notify(DEVICE_NOTIFIER_BATTERY_HEALTH, (void *)&battery.health);
#ifdef TIZEN_FEATURE_BATTERY_OVER_TEMPERATURE
			update_health(DEVICE_NOTI_ON);
#endif /* TIZEN_FEATURE_BATTERY_OVER_TEMPERATURE */
			return;
		} else if (battery.present == PRESENT_ABNORMAL) {
			update_present(DEVICE_NOTI_ON);
			return;
		}
	}
	if (battery.health == HEALTH_OVP) {
		update_ovp(DEVICE_NOTI_ON);
		return;
	}

	if (battery.charge_now != CHARGER_ABNORMAL &&
	    status == DEVICE_CHANGE_NORMAL) {
		device_notify(DEVICE_NOTIFIER_BATTERY_HEALTH, (void *)&battery.health);
		health_status_broadcast();
#ifdef TIZEN_FEATURE_BATTERY_OVER_TEMPERATURE
		update_health(DEVICE_NOTI_OFF);
#endif /* TIZEN_FEATURE_BATTERY_OVER_TEMPERATURE */
		update_ovp(DEVICE_NOTI_OFF);
		update_present(DEVICE_NOTI_OFF);
	}
}

static void check_online(void)
{
	static int old_online;
	static int old_charge_status;
	int charge_status, ret;

	if (battery.charge_status == CHARGE_STATUS_FULL)
		charge_status = CHARGE_STATUS_DISCHARGING;
	else
		charge_status = battery.charge_status;

	if (battery.online > POWER_SUPPLY_TYPE_BATTERY &&
	    old_online == VCONFKEY_SYSMAN_CHARGER_DISCONNECTED) {
		old_online = VCONFKEY_SYSMAN_CHARGER_CONNECTED;
		ret = vconf_set_int(VCONFKEY_SYSMAN_CHARGER_STATUS, old_online);
		if (ret < 0)
			_E("Failed to set vconf value for charger status: %d", vconf_get_ext_errno());
		power_supply_broadcast(CHARGER_STATUS_SIGNAL, old_online);
		check_power_supply(old_online);
		charger_state_send_system_event(CHARGE_STATUS_CONNECTED);
		if (charge_status != old_charge_status)
			charger_state_send_system_event(charge_status);

	} else if (battery.online <= POWER_SUPPLY_TYPE_BATTERY &&
	    old_online == VCONFKEY_SYSMAN_CHARGER_CONNECTED) {
		old_online = VCONFKEY_SYSMAN_CHARGER_DISCONNECTED;
		ret = vconf_set_int(VCONFKEY_SYSMAN_CHARGER_STATUS, old_online);
		if (ret < 0)
			_E("Failed to set vconf value for charger status: %d", vconf_get_ext_errno());
		power_supply_broadcast(CHARGER_STATUS_SIGNAL, old_online);
		check_power_supply(old_online);
		if (charge_status != old_charge_status)
			charger_state_send_system_event(charge_status);
		charger_state_send_system_event(CHARGE_STATUS_DISCONNECTED);

	} else {
		if (charge_status != old_charge_status)
			charger_state_send_system_event(charge_status);
	}

	old_charge_status = charge_status;
}

static void check_charge_status(const char *env_value)
{
	if (env_value == NULL)
		return;

	_D("Charge Status(%s)", env_value);

	if (strncmp(env_value, CHARGEFULL_NAME,
				sizeof(CHARGEFULL_NAME)) == 0)
		battery.charge_status = CHARGE_STATUS_FULL;
	else if (strncmp(env_value, CHARGENOW_NAME,
				sizeof(CHARGENOW_NAME)) == 0)
		battery.charge_status = CHARGE_STATUS_CHARGING;
	else if (strncmp(env_value, DISCHARGE_NAME,
				sizeof(DISCHARGE_NAME)) == 0)
		battery.charge_status = CHARGE_STATUS_DISCHARGING;
	else if (strncmp(env_value, NOTCHARGE_NAME,
				sizeof(NOTCHARGE_NAME)) == 0)
		battery.charge_status = CHARGE_STATUS_NOT_CHARGING;
	else
		battery.charge_status = CHARGE_STATUS_UNKNOWN;

	if (battery.charge_status == CHARGE_STATUS_FULL) {
		battery.charge_full = CHARGING_FULL;
		battery.charge_now = CHARGER_DISCHARGING;
	} else if (battery.charge_status == CHARGE_STATUS_CHARGING) {
		battery.charge_full = CHARGING_NOT_FULL;
		battery.charge_now = CHARGER_CHARGING;
	} else if (battery.charge_status == CHARGE_STATUS_DISCHARGING) {
		battery.charge_full = CHARGING_NOT_FULL;
		battery.charge_now = CHARGER_DISCHARGING;
	} else if (battery.charge_status == CHARGE_STATUS_NOT_CHARGING) {
		battery.charge_full = CHARGING_NOT_FULL;
		battery.charge_now = CHARGER_ABNORMAL;
	} else {
		battery.charge_full = CHARGING_NOT_FULL;
		battery.charge_now = CHARGER_DISCHARGING;
	}
}

static void check_health_status(const char *env_value)
{
	if (env_value == NULL) {
		battery.health = HEALTH_GOOD;
		return;
	}
	if (strncmp(env_value, OVERHEAT_NAME,
				sizeof(OVERHEAT_NAME)) == 0)
		battery.health = HEALTH_HIGH;
	else if (strncmp(env_value, TEMPCOLD_NAME,
				sizeof(TEMPCOLD_NAME)) == 0)
		battery.health = HEALTH_LOW;
	else if (strncmp(env_value, OVERVOLT_NAME,
				sizeof(OVERVOLT_NAME)) == 0)
		battery.health = HEALTH_OVP;
	else if (strncmp(env_value, GOOD_NAME,
				sizeof(GOOD_NAME)) == 0)
		battery.health = HEALTH_GOOD;
	else
		battery.health = HEALTH_NO_OPT;
}

static void check_online_status(const char *env_value)
{
	if (env_value == NULL)
		return;
	battery.online = atoi(env_value);
}

static void check_present_status(const char *env_value)
{
	if (env_value == NULL) {
		battery.present = PRESENT_NORMAL;
		return;
	}
	battery.present = atoi(env_value);
}

static void check_capacity_status(const char *env_value)
{
	if (env_value == NULL)
		return;
	battery.capacity = atoi(env_value);
}

static void process_power_supply(void *data)
{
	int ret;

	if (old_battery.charge_now != battery.charge_now || battery.charge_now == CHARGER_ABNORMAL) {
		ret = vconf_set_int(VCONFKEY_SYSMAN_BATTERY_CHARGE_NOW, battery.charge_now);
		if (ret < 0) {
			_E("Failed to set vconf value for battery charge now: %d", vconf_get_ext_errno());
			return;
		}
		power_supply_broadcast(CHARGE_NOW_SIGNAL, battery.charge_now);
	}

	lowbat_execute(data);
	check_online();
	if (old_battery.charge_full != battery.charge_full)
		noti_batt_full();

	old_battery.capacity = battery.capacity;
	old_battery.online = battery.online;
	old_battery.charge_status = battery.charge_status;
	old_battery.charge_now = battery.charge_now;
	old_battery.charge_full = battery.charge_full;

	check_battery_status();
	device_notify(DEVICE_NOTIFIER_POWER_SUPPLY, NULL);
	device_notify(DEVICE_NOTIFIER_BATTERY_CHARGING, &battery.charge_now);
}

static void battery_state(struct battery_info *info)
{
	if (!info)
		return;

	_I("status=%s(%s) power_source=%s(%d) capa=%d health=%s present=%d current=%d(%d) volt=%d(%d) temp=%d",
			info->status,
			battery.charge_now == CHARGER_CHARGING ? "Charging"
				: (battery.charge_now == CHARGER_DISCHARGING ? "Discharging" : "Abnormal"),
			info->power_source,
			info->online,
			info->capacity,
			info->health,
			info->present,
			info->current_now,
			info->current_average,
			info->voltage_now,
			info->voltage_average,
			info->temperature);
}

static void battery_changed(struct battery_info *info, void *data)
{
	int ret;

	if (!info)
		return;

	if (info->status) {
		snprintf(battery.status_s, sizeof(battery.status_s),
				"%s", info->status);
		check_charge_status(info->status);
	} else
		battery.status_s[0] = '\0';

	if (info->health) {
		snprintf(battery.health_s, sizeof(battery.health_s),
				"%s", info->health);
		check_health_status(info->health);
	} else
		battery.health_s[0] = '\0';

	if (info->power_source)
		snprintf(battery.power_source_s, sizeof(battery.power_source_s),
				"%s", info->power_source);
	else
		battery.power_source_s[0] = '\0';

	battery.online = info->online;
	battery.present = info->present;
	battery.capacity = info->capacity;
	battery.current_now = info->current_now;
	battery.current_average = info->current_average;
	battery.voltage_now = info->voltage_now;
	battery.voltage_average = info->voltage_average;
	battery.temperature = info->temperature;

	battery_state(info);

	ret = booting_done(NULL);
	if (ret) {
		if (battery.online > POWER_SUPPLY_TYPE_BATTERY)
			power_supply_noti(DEVICE_NOTI_BATT_CHARGE, DEVICE_NOTI_ON);
		else
			power_supply_noti(DEVICE_NOTI_BATT_CHARGE, DEVICE_NOTI_OFF);
	}

	process_power_supply(&battery.capacity);

}

static void battery_get_capacity(struct battery_info *info, void *data)
{
	int *capa = data;

	if (info && info->capacity >= 0)
		*capa = info->capacity;
}

static void power_supply_status_init(void)
{
	static int charge_now = -1;
	static int charge_full = -1;
	static int capacity = -1;
	int pct;
	int r;

	pct = -1;
	r = battery_dev->get_current_state(battery_get_capacity, &pct);
	if (r < 0 || pct < 0) {
		_E("Failed to get battery capacity. capa=%d: %d", pct, r);
		return;
	}

	battery.capacity = pct;
	battery.health = HEALTH_GOOD;
	battery.present = PRESENT_NORMAL;

	if (charge_now == battery.charge_now &&
	    charge_full == battery.charge_full &&
	    capacity == battery.capacity)
		return;

	_I("charging=%d full=%d capacity=%d", battery.charge_now, battery.charge_full, battery.capacity);

	if (charge_now != battery.charge_now) {
		r = vconf_set_int(VCONFKEY_SYSMAN_BATTERY_CHARGE_NOW, battery.charge_now);
		if (r < 0)
			_E("Failed to set vconf value for battery charge now: %d", vconf_get_ext_errno());
		power_supply_broadcast(CHARGE_NOW_SIGNAL, battery.charge_now);
	}
	if (capacity != battery.capacity) {
		r = vconf_set_int(VCONFKEY_SYSMAN_BATTERY_CAPACITY, battery.capacity);
		if (r < 0)
			_E("Failed to set vconf value for battery capacity: %d", vconf_get_ext_errno());
		power_supply_broadcast(CHARGE_CAPACITY_SIGNAL, battery.capacity);
	}

	charge_now = battery.charge_now;
	charge_full = battery.charge_full;
	capacity =  battery.capacity;
}

static gboolean power_supply_update(void *data)
{
	power_supply_status_init();
	return G_SOURCE_CONTINUE;
}

static void power_supply_timer_start(void)
{
	_D("Battery init timer during booting.");
	power_timer = g_timeout_add(BATTERY_CHECK_TIMER_INTERVAL,
				power_supply_update, NULL);
	if (power_timer == 0)
		_E("Failed to add battery init timer during booting.");
}

static void power_supply_timer_stop(void)
{
	_D("Battery init timer during booting.");
	if (!power_timer)
		return;
	g_source_remove(power_timer);
	power_timer = 0;
}

static GVariant *dbus_get_charger_status(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret, status;

	ret = vconf_get_int(VCONFKEY_SYSMAN_CHARGER_STATUS, &status);
	if (ret < 0) {
		_E("Failed to get vconf value for charger status: %d", vconf_get_ext_errno());
		status = -EIO;
	}
	return g_variant_new("(i)", status);
}

static GVariant *dbus_get_charge_now(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret;

	ret = battery.charge_now;

	return g_variant_new("(i)", ret);
}

static GVariant *dbus_get_charge_level(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret, status;

	ret = vconf_get_int(VCONFKEY_SYSMAN_BATTERY_STATUS_LOW, &status);
	if (ret < 0) {
		_E("Failed to get vconf value for battery status low: %d", vconf_get_ext_errno());
		status = -EIO;
	}

	return g_variant_new("(i)", status);
}

static GVariant *dbus_get_percent(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret;

	ret = battery.capacity;

	return g_variant_new("(i)", ret);
}

static GVariant *dbus_get_percent_raw(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret;

	ret = -ENOTSUP;

	return g_variant_new("(i)", ret);
}

static GVariant *dbus_is_full(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret;

	ret = battery.charge_full;

	return g_variant_new("(i)", ret);
}

static GVariant *dbus_get_health(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret;

	ret = battery.health;

	return g_variant_new("(i)", ret);
}

static GVariant *dbus_power_supply_handler(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	pid_t pid;
	int ret = 0;
	int argc = 0;
	char *type_str;
	char *argv[5];

	g_variant_get(param, "(sisssss)", &type_str,
					&argc,
					&argv[0],
					&argv[1],
					&argv[2],
					&argv[3],
					&argv[4]);

	if (argc < 0) {
		_E("Message is invalid.");
		ret = -EINVAL;
		goto out;
	}

	pid = dbus_connection_get_sender_pid(conn, sender);
	if (kill(pid, 0) == -1) {
		_E("Process(%d) does not exist, dbus ignored.", pid);
		ret = -ESRCH;
		goto out;
	}
	check_capacity_status(argv[0]);
	check_charge_status(argv[1]);
	check_health_status(argv[2]);
	check_online_status(argv[3]);
	check_present_status(argv[4]);
	_I("capa=%d full=%d now=%d health=%d online=%d present=%d temp=%d",
		battery.capacity,
		battery.charge_full,
		battery.charge_now,
		battery.health,
		battery.online,
		battery.present,
		battery.temperature);

	if (battery.online > POWER_SUPPLY_TYPE_BATTERY)
		power_supply_noti(DEVICE_NOTI_BATT_CHARGE, DEVICE_NOTI_ON);
	else
		power_supply_noti(DEVICE_NOTI_BATT_CHARGE, DEVICE_NOTI_OFF);

	process_power_supply(&battery.capacity);
out:
	g_free(type_str);
	g_free(argv[0]);
	g_free(argv[1]);
	g_free(argv[2]);
	g_free(argv[3]);
	g_free(argv[4]);

	return g_variant_new("(i)", ret);
}

static void battery_get_info(struct battery_info *info, void *data)
{
	struct battery_info *bat = data;

	if (!info || !bat)
		return;

	bat->status = strdup(info->status);
	bat->health = strdup(info->health);
	bat->power_source = strdup(info->power_source);
	bat->online = info->online;
	bat->present = info->present;
	bat->capacity = info->capacity;
	bat->current_now = info->current_now;
	bat->current_average = info->current_average;
	bat->voltage_now = info->voltage_now;
	bat->voltage_average = info->voltage_average;
	bat->temperature = info->temperature;
}

static GVariant *dbus_get_battery_info(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret;
	struct battery_info info = { 0, };

	ret = battery_dev->get_current_state(battery_get_info, &info);
	if (ret < 0)
		_E("Failed to get battery info: %d", ret);

	battery_changed(&info, NULL);
	free(info.status);
	free(info.health);
	free(info.power_source);

	return g_variant_new("(isssiiiiiiii)", ret,
					battery.status_s,
					battery.health_s,
					battery.power_source_s,
					battery.online,
					battery.present,
					battery.capacity,
					battery.current_now,
					battery.current_average,
					battery.voltage_now,
					battery.voltage_average,
					battery.temperature);
}

static const dbus_method_s dbus_methods[] = {
	{ CHARGER_STATUS_SIGNAL,      NULL, "i", dbus_get_charger_status },
	{ CHARGE_NOW_SIGNAL,          NULL, "i", dbus_get_charge_now },
	{ CHARGE_LEVEL_SIGNAL,        NULL, "i", dbus_get_charge_level },
	{ CHARGE_CAPACITY_SIGNAL,     NULL, "i", dbus_get_percent },
	{ CHARGE_CAPACITY_LAW_SIGNAL, NULL, "i", dbus_get_percent_raw },
	{ CHARGE_FULL_SIGNAL,         NULL, "i", dbus_is_full },
	{ CHARGE_HEALTH_SIGNAL,       NULL, "i", dbus_get_health },
	{ POWER_SUBSYSTEM,       "sisssss", "i", dbus_power_supply_handler },
	{ "GetBatteryInfo",           NULL, "isssiiiiiiii", dbus_get_battery_info },
	/* Add methods here */
};

static const dbus_interface_u dbus_interface = {
	.oh = NULL,
	.name = DEVICED_INTERFACE_BATTERY,
	.methods = dbus_methods,
	.nr_methods = ARRAY_SIZE(dbus_methods),
};

static int booting_done(void *data)
{
	static int done;

	if (data == NULL)
		return done;
	done = *(int *)data;
	if (done == 0)
		return done;

	_I("Booting done.");

	power_supply_timer_stop();

	/* for simple noti change cb */
	power_supply_status_init();
	process_power_supply(NULL);

	return done;
}

static int display_changed(void *data)
{
	if (battery.charge_now != CHARGER_ABNORMAL)
		return 0;
	if (battery.health != HEALTH_DEAD && battery.present != PRESENT_ABNORMAL)
		return 0;

	if (disp_plgn.pm_lock_internal)
		disp_plgn.pm_lock_internal(INTERNAL_LOCK_POPUP, LCD_DIM, STAY_CUR_STATE, 0);
	return 0;
}

static int power_supply_probe(void *data)
{
	struct hw_info *info;
	int ret, r;

	if (battery_dev)
		return 0;

	ret = hw_get_info(BATTERY_HARDWARE_DEVICE_ID,
			(const struct hw_info **)&info);
	if (ret < 0) {
		_E("Failed to load battery shared library: %d", ret);
		goto out;
	}

	if (!info->open) {
		_E("Failed to open battery device: open(NULL)");
		ret = -ENODEV;
		goto out;
	}

	ret = info->open(info, NULL, (struct hw_common**)&battery_dev);
	if (ret < 0) {
		_E("Failed to get battery device structure: %d", ret);
		goto out;
	}

	if (!battery_dev || !battery_dev->get_current_state) {
		_E("get_current_state() is not supported by the Battery HAL.");
		ret = -ENODEV;
		goto out;
	}

	_I("Battery device structure load success.");
	return 0;

out:
	/**
	 * Set battery vconf as -ENOTSUP
	 * These vconf key used by runtime-info and capi-system-device.
	 */
	r = vconf_set_int(VCONFKEY_SYSMAN_CHARGER_STATUS, -ENOTSUP);
	if (r < 0)
		_E("Failed to set vconf value for charger status: %d", vconf_get_ext_errno());
	r = vconf_set_int(VCONFKEY_SYSMAN_BATTERY_CHARGE_NOW, -ENOTSUP);
	if (r < 0)
		_E("Failed to set vconf value for battery charge now: %d", vconf_get_ext_errno());
	r = vconf_set_int(VCONFKEY_SYSMAN_BATTERY_LEVEL_STATUS, -ENOTSUP);
	if (r < 0)
		_E("Failed to set vconf value for battery level status: %d", vconf_get_ext_errno());
	_I("There is no battery device: %d", ret);
	return -ENODEV;
}

static void add_power_supply_handler(void)
{
	if (battery_dev->register_changed_event)
		battery_dev->register_changed_event(battery_changed, NULL);

	/* already checked in power_supply_probe() */
	battery_dev->get_current_state(battery_changed, NULL);
}

static void remove_power_supply_handler(void)
{
	if (battery_dev->unregister_changed_event)
		battery_dev->unregister_changed_event(battery_changed);
}

static int event_handler_state_changed(void *data)
{
	static device_notifier_state_e old = DEVICE_NOTIFIER_STATE_STOP;
	device_notifier_state_e state = *(device_notifier_state_e *)data;

	if (old == state)
		return 0;

	old = state;

	if (state == DEVICE_NOTIFIER_STATE_START)
		add_power_supply_handler();
	else if (state == DEVICE_NOTIFIER_STATE_STOP)
		remove_power_supply_handler();

	return 0;
}

static void power_supply_init(void *data)
{
	int ret;
	device_notifier_state_e state = DEVICE_NOTIFIER_STATE_START;

	battery.capacity = -1;

	event_handler_state_changed((void *)&state);

	/* process check battery timer until booting done */
	power_supply_timer_start();

	register_notifier(DEVICE_NOTIFIER_BOOTING_DONE, booting_done);
	register_notifier(DEVICE_NOTIFIER_LCD, display_changed);
	register_notifier(DEVICE_NOTIFIER_EVENT_HANDLER, event_handler_state_changed);

	ret = dbus_handle_add_dbus_object(NULL, DEVICED_PATH_BATTERY, &dbus_interface);
	if (ret < 0)
		_E("Failed to init dbus method: %d", ret);


#ifdef TIZEN_FEATURE_BATTERY_OVER_TEMPERATURE
	ret = subscribe_dbus_signal(NULL, DEVICED_PATH_SYSNOTI,
		DEVICED_INTERFACE_SYSNOTI, SIGNAL_CHARGEERR_RESPONSE, abnormal_popup_dbus_signal_handler, NULL, NULL);
	if (ret <= 0)
		_E("Failed to init dbus signal: %d", ret);
#endif /* TIZEN_FEATURE_BATTERY_OVER_TEMPERATURE */
}

static void power_supply_exit(void *data)
{
	device_notifier_state_e state = DEVICE_NOTIFIER_STATE_STOP;

	unregister_notifier(DEVICE_NOTIFIER_BOOTING_DONE, booting_done);
	unregister_notifier(DEVICE_NOTIFIER_LCD, display_changed);
	unregister_notifier(DEVICE_NOTIFIER_EVENT_HANDLER, event_handler_state_changed);

	event_handler_state_changed((void *)&state);
}

static const struct device_ops power_supply_ops = {
	.name     = "power_supply",
	.probe    = power_supply_probe,
	.init     = power_supply_init,
	.exit     = power_supply_exit,
};

DEVICE_OPS_REGISTER(&power_supply_ops)
