/*
 * deviced
 *
 * Copyright (c) 2014 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <fcntl.h>

#include "core/log.h"
#include "core/list.h"
#include "core/common.h"
#include "core/devices.h"
#include <libsyscommon/dbus-system.h>

#define SERVICE_NAME		"deviced"
#define DUMP_SIGNAL		"Dump"
#define DUMP_START_SIGNAL	"Start"
#define DUMP_FINISH_SIGNAL	"Finish"
#define DEFAULT_DUMP_PATH	"/var/log"

static void send_dump_signal(char *signal)
{
	int ret;
	pid_t pid;

	pid = getpid();

	ret = dbus_handle_broadcast_dbus_signal_var(DUMP_SERVICE_OBJECT_PATH,
			DUMP_SERVICE_INTERFACE_NAME, signal,
			g_variant_new("(i)", pid));
	if (ret < 0)
		_E("Failed to send dbus signal(%s)", signal);
}

static void dump_all_devices(int mode, char *path)
{
	dd_list *head = get_device_list_head();
	dd_list *elem;
	FILE *fp = NULL;
	char fname[PATH_MAX];
	const struct device_ops *dev;

	/* send start signal to dump service */
	send_dump_signal(DUMP_START_SIGNAL);

	/* open saved file */
	snprintf(fname, PATH_MAX, "%s/%s.log", path, SERVICE_NAME);

	if (path)
		fp = fopen(fname, "w+");
	if (!fp)
		_I("Failed to open '%s', print to DLOG.", fname);

	/* save dump each device ops */
	DD_LIST_FOREACH(head, elem, dev) {
		if (dev->dump) {
			_D("[%s] Get dump.", dev->name);
			LOG_DUMP(fp, "\n==== %s\n\n", dev->name);
			dev->dump(fp, mode, dev->dump_data);
		}
	}

	/* close file */
	if (fp)
		fclose(fp);

	/* send finish signal to dump service */
	send_dump_signal(DUMP_FINISH_SIGNAL);

	_D("Dump(%s:%d) is saved!", fname, mode);
}

static void dump_signal_handler(GDBusConnection  *conn,
	const gchar      *sender,
	const gchar      *path,
	const gchar      *iface,
	const gchar      *name,
	GVariant         *param,
	gpointer          data)

{
	int mode = 0;
	char *log_path = NULL;

	g_variant_get(param, "(is)", &mode, &log_path);
	_D("mode=%d path=%s", mode, log_path);

	dump_all_devices(mode, log_path);
	g_free(log_path);
}

static void dump_init(void *data)
{
	int ret;

	/* register dump signal */
	ret = subscribe_dbus_signal(NULL,
				DUMP_SERVICE_OBJECT_PATH,
				DUMP_SERVICE_INTERFACE_NAME,
				DUMP_SIGNAL,
				dump_signal_handler,
				NULL, NULL);
	if (ret <= 0)
		_E("Failed to register signal handler: %d", ret);
}

static const struct device_ops dump_device_ops = {
	.name     = "dump",
	.init     = dump_init,
};

DEVICE_OPS_REGISTER(&dump_device_ops)

