/*
 * deviced
 *
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <stdbool.h>

#include <hw/thermal.h>
#include <libsyscommon/dbus-system.h>

#include "apps/apps.h"
#include "core/devices.h"
#include "core/log.h"
#include "core/device-notifier.h"
#include "core/devices.h"
#include "thermal.h"

static struct thermal_device *thermal_dev;

static int booting_done(void *data)
{
	static int done;

	if (data == NULL)
		return done;

	done = *(int *)data;
	if (done == 0)
		return done;

	_I("Booting done.");
	return done;
}

static GVariant *dbus_get_temperature(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	struct thermal_info info;
	int index, ret = 0;

	g_variant_get(param, "(i)", &index);

	if (!thermal_dev || !thermal_dev->get_info) {
		ret = -ENODEV;
		goto exit;
	}

	ret = thermal_dev->get_info((device_thermal_e)index, &info);
	if (!ret)
		ret = info.temp;

exit:
	return g_variant_new("(i)", ret);
}

static const dbus_method_s dbus_methods[] = {
	{ "GetTemperature", "i", "i", dbus_get_temperature},
	/* Add methods here */
};

static const dbus_interface_u dbus_interface = {
	.oh = NULL,
	.name = DEVICED_INTERFACE_TEMPERATURE,
	.methods = dbus_methods,
	.nr_methods = ARRAY_SIZE(dbus_methods),
};

static void thermal_init(void *data)
{
	int ret;

	ret = register_notifier(DEVICE_NOTIFIER_BOOTING_DONE, booting_done);
	if (ret < 0)
		_E("Failed to register booting done notifier.");

	ret = dbus_handle_add_dbus_object(NULL, DEVICED_PATH_TEMPERATURE, &dbus_interface);
	if (ret < 0)
		_E("Failed to init dbus method: %d", ret);
}

static void thermal_exit(void *data)
{
	int ret;

	ret = unregister_notifier(DEVICE_NOTIFIER_BOOTING_DONE, booting_done);
	if (ret < 0)
		_E("Failed to unregister booting done notifier.");
}

static int thermal_probe(void *data)
{
	struct hw_info *info;
	int ret;

	if (thermal_dev)
		return 0;

	ret = hw_get_info(THERMAL_HARDWARE_DEVICE_ID,
			(const struct hw_info **)&info);
	if (ret < 0) {
		_E("There is no HAL for Thermal.");
		return ret;
	}

	if (!info->open) {
		_E("Failed to open thermal device: open(NULL)");
		return -ENODEV;
	}

	ret = info->open(info, NULL, (struct hw_common**)&thermal_dev);
	if (ret < 0) {
		_E("Failed to get thermal device structure: %d", ret);
		return ret;
	}

	if (!thermal_dev ||
		!thermal_dev->get_info ||
		!thermal_dev->register_changed_event ||
		!thermal_dev->unregister_changed_event) {
		_E("Thermal HAL does not support event handler.");
		return -ENODEV;
	}

	_I("Thermal device structure load success.");
	return 0;
}

static const struct device_ops thermal_ops = {
	.name     = "thermal",
	.probe    = thermal_probe,
	.init     = thermal_init,
	.exit     = thermal_exit,
};

DEVICE_OPS_REGISTER(&thermal_ops)
